@@ -1,5 +1,5 @@
 /*!
-  Copyright (C) 2018-2020 Arthur de Jong
+  Copyright (C) 2018-2021 Arthur de Jong
 
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
@@ -70,7 +70,7 @@ $(document).ready(function () {
     start.round10Minutes('floor')
     end.round10Minutes('ceil')
     // update the date range picker
-    var daterangepicker = $('#reportrange').data('daterangepicker')
+    const daterangepicker = $('#reportrange').data('daterangepicker')
     daterangepicker.setStartDate(start)
     daterangepicker.setEndDate(end)
     // ensure start and end are strings
@@ -105,32 +105,29 @@ $(document).ready(function () {
     alwaysShowCalendars: true,
     ranges: {
       Today: [moment().subtract(1, 'days').round10Minutes(), moment().add(1, 'hour').round10Minutes('ceil')],
-      Yesterday: [moment().subtract(1, 'days').startOf('day'), moment().subtract(1, 'days').endOf('day').round10Minutes('ceil')],
-      'Last 7 days': [moment().subtract(6, 'days').startOf('day'), moment().endOf('day').round10Minutes('ceil')],
-      'Last 30 days': [moment().subtract(29, 'days').startOf('day'), moment().endOf('day').round10Minutes('ceil')],
-      'This month': [moment().startOf('month'), moment().endOf('month').round10Minutes('ceil')],
-      'Last month': [moment().subtract(1, 'month').startOf('month'), moment().subtract(1, 'month').endOf('month').round10Minutes('ceil')],
+      'This week': [moment().subtract(6, 'days').startOf('day'), moment().endOf('day').round10Minutes('ceil')],
+      'This month': [moment().subtract(32, 'days').startOf('day'), moment().endOf('day').round10Minutes('ceil')],
       'This year': [moment().subtract(365, 'days').startOf('month'), moment().endOf('month').round10Minutes('ceil')]
     }
   }, setDateRange)
 
   try {
     // restore the previously saved date range
-    var data = JSON.parse(localStorage.getItem('dateRange'))
+    const data = JSON.parse(localStorage.getItem('dateRange'))
     setDateRange(data.start, data.end)
   } catch (error) {
     // set a default date range
     setDateRange(moment().subtract(2, 'days'), moment().add(1, 'hour').round10Minutes('ceil'))
   }
 
-  var defaultColors = [
+  const defaultColors = [
     '#00cc00', '#0066b3', '#ff8000', '#dbc300', '#330099', '#990099',
     '#bce617', '#ff0000', '#808080', '#008f00', '#00487d', '#b35a00',
     '#b38f00', '#6b006b', '#8fb300', '#b30000', '#bebebe', '#80ff80',
     '#80c9ff', '#ffc080', '#ffe680', '#aa80ff', '#ee00cc', '#ff8080',
     '#666600', '#ffbfff', '#00ffcc', '#cc6699', '#999900']
 
-  var baseLayout = {
+  const baseLayout = {
     margin: {l: 48, t: 0, r: 8, b: 32},
     autosize: true,
     showlegend: false,
@@ -165,7 +162,7 @@ $(document).ready(function () {
     datarevision: 1
   }
 
-  var config = {
+  const config = {
     showLink: false,
     displaylogo: false,
     autosizable: true,
@@ -184,23 +181,24 @@ $(document).ready(function () {
   }
 
   // whether new data should be loaded
-  var updatedata = false
+  let updatedata = false
 
   // update the legend
   function updateLegend(plot) {
-    var [minx, maxx] = plot.layout.xaxis.range
+    const [minx, maxx] = plot.layout.xaxis.range
     Object.keys(plot.legendbyfield).forEach(function (field) {
       // calculate minimum
-      var mintrace = plot.tracebyfield[field + '.min']
-      var minvalue = Math.min.apply(null, mintrace.y.filter(function (el, idx) {
-        var x = mintrace.x[idx]
+      const mintrace = plot.tracebyfield[field + '.min']
+      const minvalue = Math.min.apply(null, mintrace.y.filter(function (el, idx) {
+        const x = mintrace.x[idx]
         return x >= minx && x <= maxx
       }))
       // calculate average
-      var avgtrace = plot.tracebyfield[field]
+      const avgtrace = plot.tracebyfield[field]
+      let avgvalue
       if (avgtrace.y.length) {
-        var avgvalue = avgtrace.y.map(function (current, idx) {
-          var x = avgtrace.x[idx]
+        avgvalue = avgtrace.y.map(function (current, idx) {
+          const x = avgtrace.x[idx]
           if (idx > 0 && x >= minx && x <= maxx) {
             return [current, Date.parse(x) - Date.parse(avgtrace.x[idx - 1])]
           } else {
@@ -214,13 +212,13 @@ $(document).ready(function () {
         avgvalue = undefined
       }
       // calculate maximum
-      var maxtrace = plot.tracebyfield[field + '.max']
-      var maxvalue = Math.max.apply(null, maxtrace.y.filter(function (el, idx) {
-        var x = maxtrace.x[idx]
+      const maxtrace = plot.tracebyfield[field + '.max']
+      const maxvalue = Math.max.apply(null, maxtrace.y.filter(function (el, idx) {
+        const x = maxtrace.x[idx]
         return x >= minx && x <= maxx
       }))
       // update legend
-      var columns = $(plot.legendbyfield[field]).find('td')
+      const columns = $(plot.legendbyfield[field]).find('td')
       columns[2].textContent = (isNaN(minvalue) || !isFinite(minvalue)) ? '-' : Plotly.d3.format('.4s')(minvalue)
       columns[3].textContent = (isNaN(avgvalue) || !isFinite(avgvalue)) ? '-' : Plotly.d3.format('.4s')(avgvalue)
       columns[4].textContent = (isNaN(maxvalue) || !isFinite(maxvalue)) ? '-' : Plotly.d3.format('.4s')(maxvalue)
@@ -230,7 +228,7 @@ $(document).ready(function () {
   // load graph data into the plot
   function loadGraph(plot, legend, graph) {
     // prepare the graph configuration
-    var layout = JSON.parse(JSON.stringify(baseLayout))
+    const layout = JSON.parse(JSON.stringify(baseLayout))
     if (graph.graph_vlabel) {
       layout.yaxis.title = graph.graph_vlabel
     }
@@ -239,23 +237,23 @@ $(document).ready(function () {
       layout.yaxis.exponentformat = 'E'
     }
     // get x axis zoom from date range selector
-    var daterangepicker = $('#reportrange').data('daterangepicker')
+    const daterangepicker = $('#reportrange').data('daterangepicker')
     layout.xaxis.range = [
       daterangepicker.startDate.format('YYYY-MM-DD HH:mm'),
       daterangepicker.endDate.format('YYYY-MM-DD HH:mm')]
     // prepare the data series configuration
-    var traces = []
-    var tracebyfield = {}
+    const traces = []
+    const tracebyfield = {}
     plot.tracebyfield = tracebyfield
-    var stackgroup = 0
-    for (var i = 0; i < graph.fields.length; i++) {
-      var field = graph.fields[i]
-      var color = field.colour ? '#' + field.colour : defaultColors[i % defaultColors.length]
+    let stackgroup = 0
+    for (let i = 0; i < graph.fields.length; i++) {
+      const field = graph.fields[i]
+      const color = field.colour ? '#' + field.colour : defaultColors[i % defaultColors.length]
       if (field.draw === 'AREA' || field.draw === 'STACK' || field.draw === 'AREASTACK') {
         if (!field.draw.match(/STACK/) && (!graph.fields[i + 1] || graph.fields[i + 1].draw.match(/STACK/))) {
           stackgroup += 1
         }
-        var trace = {
+        const trace = {
           field_name: field.name,
           name: field.label || field.name,
           info: field.info || '',
@@ -269,20 +267,20 @@ $(document).ready(function () {
         tracebyfield[field.name + '.min'] = {}
         tracebyfield[field.name + '.max'] = {}
       } else if (field.draw) {
-        trace = {
+        const trace = {
           field_name: field.name,
           name: field.label || field.name,
           info: field.info || '',
           line: {color: color},
           hoverlabel: {bgcolor: color + 'c0'}
         }
-        var minTrace = {
+        const minTrace = {
           field_name: field.name,
           showlegend: false,
           hoverinfo: 'skip',
           line: {width: 0}
         }
-        var maxTrace = {
+        const maxTrace = {
           field_name: field.name,
           showlegend: false,
           hoverinfo: 'skip',
@@ -309,7 +307,7 @@ $(document).ready(function () {
     plot.legendbyfield = {}
     traces.slice().reverse().forEach(function (trace) {
       if (trace.showlegend !== false) {
-        var legendrow = $('<tr></tr>')
+        const legendrow = $('<tr></tr>')
         legendrow.append($('<td style="width: 30px;"><svg height="10" width="20"><line x1="0" y1="5" x2="20" y2="5"></svg></td>'))
         legendrow.append($('<td><span></span></td>'))
         legendrow.append($('<td></td>'))
@@ -326,7 +324,7 @@ $(document).ready(function () {
         // handle showing/hiding the trace
         legendrow.click(function () {
           if (plot.data) {
-            var visible = (trace.visible === false)
+            const visible = (trace.visible === false)
             $(this).css('opacity', visible ? 1 : 0.2)
             plot.data.forEach(function (t) {
               if (t.field_name === trace.field_name) {
@@ -337,40 +335,38 @@ $(document).ready(function () {
             Plotly.redraw(plot)
           }
         })
-        // highlight the trace by lowering the opacity of the others
+        // highlight the trace by lowering the opacity of the others traces
         legendrow.mouseover(function () {
           if (plot.data) {
-            var vals = plot.data.map(t => t.field_name === trace.field_name ? 1 : 0.1)
-            Plotly.restyle(plot, 'opacity', vals)
-            vals = plot.data.map(function (t) {
+            Plotly.restyle(plot, 'opacity', plot.data.map(
+              t => t.field_name === trace.field_name ? 1 : 0.1))
+            Plotly.restyle(plot, 'fillcolor', plot.data.map(function (t) {
               if (t.showlegend === false) {
-                return t.fillcolor
+                return (t.fillcolor || '#ffffff').substring(0, 7) + (t.field_name === trace.field_name ? '25' : '05')
               }
               return (t.fillcolor || '#ffffff').substring(0, 7) + (t.field_name === trace.field_name ? 'ff' : '30')
-            })
-            Plotly.restyle(plot, 'fillcolor', vals)
+            }))
           }
         })
       }
     })
-    // reset opacity after exiting the legend
+    // restore opacity after exiting the legend
     legend.mouseout(function () {
       if (plot.data) {
         Plotly.restyle(plot, 'opacity', plot.data.map(t => 1))
-        var vals = plot.data.map(function (t) {
+        Plotly.restyle(plot, 'fillcolor', plot.data.map(function (t) {
           if (t.showlegend === false) {
-            return t.fillcolor
+            return (t.fillcolor || '#ffffff').substring(0, 7) + '20'
           }
           return (t.fillcolor || '#ffffff').substring(0, 7) + 'c0'
-        })
-        Plotly.restyle(plot, 'fillcolor', vals)
+        }))
       }
     })
     // fetch the data and plot it
     // TODO: probably skip this initial load???
     Plotly.d3.csv('data/' + graph.name, function (data) {
-      for (var i = 0; i < data.length; i++) {
-        var row = data[i]
+      for (let i = 0; i < data.length; i++) {
+        const row = data[i]
         Object.keys(tracebyfield).forEach(function (field) {
           tracebyfield[field].x.push(row.time)
           tracebyfield[field].y.push(Number(row[field]))
@@ -401,66 +397,25 @@ $(document).ready(function () {
         updatedata = false
         // go over all plots
         $('.myplot').each(function () {
-          var plot = this
+          const plot = this
           if (plot.layout) {
             // range of the x axis
-            var [amin, amax] = plot.layout.xaxis.range
-            // range of the currently loaded data
-            var dmin = plot.data.map(t => t.x[0]).reduce((a, c) => a < c ? a : c)
-            var dmax = plot.data.map(t => t.x[t.x.length - 1]).reduce((a, c) => a > c ? a : c)
-            // range that we have marked as loaded
-            // (to avoid retrying to load data that isn't there)
-            if (!plot.lmin) {
-              plot.lmin = dmin
-            }
-            if (!plot.lmax) {
-              plot.lmax = dmax
-            }
-            // see if we need to load data before the currently loaded range
-            if (amin < plot.lmin) {
-              plot.lmin = amin
-              var url = 'data/' + plot.graph.name + '?start=' + amin.substring(0, 16) + '&end=' + dmin.substring(0, 16)
-              Plotly.d3.csv(url, function (data) {
-                // prepend new data
-                if (data) {
-                  for (var i = data.length - 1; i >= 0; i--) {
-                    var row = data[i]
-                    var time = row.time
-                    Object.entries(plot.tracebyfield).forEach(function ([field, trace]) {
-                      if (time < trace.x[0]) {
-                        trace.x.splice(0, 0, time)
-                        trace.y.splice(0, 0, Number(row[field]))
-                      }
-                    })
-                  }
-                  plot.layout.datarevision += 1
-                  Plotly.react(plot, plot.data, plot.layout)
-                }
+            const [amin, amax] = plot.layout.xaxis.range
+            const url = 'data/' + plot.graph.name + '?start=' + amin.substring(0, 16).replace(' ', 'T') + '&end=' + amax.substring(0, 16).replace(' ', 'T')
+            Plotly.d3.csv(url, function (data) {
+              Object.keys(plot.tracebyfield).forEach(function (field) {
+                plot.tracebyfield[field].x = []
+                plot.tracebyfield[field].y = []
               })
-            }
-            // see if we need to load data past the currently loaded range
-            if (amax > plot.lmax) {
-              plot.lmax = amax
-              // load data from dmax to amax and append
-              url = 'data/' + plot.graph.name + '?start=' + dmax.substring(0, 16) + '&end=' + amax.substring(0, 16)
-              Plotly.d3.csv(url, function (data) {
-                // append new data
-                if (data) {
-                  for (var i = 0; i < data.length; i++) {
-                    var row = data[i]
-                    var time = row.time
-                    Object.entries(plot.tracebyfield).forEach(function ([field, trace]) {
-                      if (time > trace.x[trace.x.length - 1]) {
-                        trace.x.push(time)
-                        trace.y.push(Number(row[field]))
-                      }
-                    })
-                  }
-                  plot.layout.datarevision += 1
-                  Plotly.react(plot, plot.data, plot.layout)
-                }
+              data.forEach(function (row) {
+                Object.keys(plot.tracebyfield).forEach(function (field) {
+                  plot.tracebyfield[field].x.push(row.time)
+                  plot.tracebyfield[field].y.push(Number(row[field]))
+                })
               })
-            }
+              plot.layout.datarevision += 1
+              Plotly.react(plot, plot.data, plot.layout)
+            })
           }
         })
       }
@@ -473,17 +428,20 @@ $(document).ready(function () {
   // every minute check if there is any new data
   function checkNewData() {
     setTimeout(checkNewData, 60000)
-    $('.myplot').each(function () {
-      this.lmax = undefined
+    $('.myplot').each(function (index, plot) {
+      // if any plot has incomplete data, reload the data
+      if (plot.layout) {
+        const dmax = plot.data.map(t => t.x[t.x.length - 1]).reduce((a, c) => a > c ? a : c)
+        if (dmax < plot.layout.xaxis.range[1]) { updatedata = true }
+      }
     })
-    updatedata = true
   }
   setTimeout(checkNewData, 60000)
 
   function addGraph(graph, size = 'sm') {
-    var clone = $('#template>:first-child').clone()
-    var plot = clone.find('.myplot')[0]
-    var legend = clone.find('.mylegend')
+    const clone = $('#template>:first-child').clone()
+    const plot = clone.find('.myplot')[0]
+    const legend = clone.find('.mylegend')
     plot.graph = graph
     // update graph title
     clone.find('.graphtitle').text(graph.host + ' / ')
@@ -543,10 +501,10 @@ $(document).ready(function () {
   // update the select widget to be able to list the known graphs
   function updateSelect(graphs) {
     // make lists of groups, hosts and categories
-    var hosts = {}
-    var categories = []
-    for (var graph in graphs) {
-      var parts = graph.split('/')
+    const hosts = {}
+    const categories = []
+    for (const graph in graphs) {
+      const parts = graph.split('/')
       if (!hosts[parts[0]]) {
         hosts[parts[0]] = []
       }
@@ -559,7 +517,7 @@ $(document).ready(function () {
     }
     // update options in host selector
     Object.keys(hosts).sort().forEach(function (group) {
-      var groupElement = $('<optgroup></optgroup>').attr('label', group)
+      const groupElement = $('<optgroup></optgroup>').attr('label', group)
       hosts[group].sort().forEach(function (host) {
         groupElement.append($('<option></option>').attr('value', group + '/' + host).text(host))
       })
@@ -571,9 +529,9 @@ $(document).ready(function () {
     })
     // handler for updating the choices in the graph select
     function updateGraphList() {
-      var search = $('#graphfilter').val().toLowerCase().split(' ')
-      var hostFilter = $('#hostselect').val()
-      var categoryFilter = $('#categoryselect').val()
+      const search = $('#graphfilter').val().toLowerCase().split(' ')
+      const hostFilter = $('#hostselect').val()
+      const categoryFilter = $('#categoryselect').val()
       $('#graphselect').empty()
       Object.keys(graphs).sort().forEach(function (graph) {
         if (hostFilter && !graph.startsWith(hostFilter + '/')) {
@@ -582,12 +540,12 @@ $(document).ready(function () {
         if (categoryFilter && graphs[graph].category !== categoryFilter) {
           return
         }
-        var descripton = (graph + ' ' + graphs[graph].graph_title + ' ' + graphs[graph].category).toLowerCase()
+        const descripton = (graph + ' ' + graphs[graph].graph_title + ' ' + graphs[graph].category).toLowerCase()
         if (search.some(x => !descripton.includes(x))) {
           return
         }
-        var title = graphs[graph].graph_title || graph.split('/')[2]
-        var graphelement = $('<a href="#" class="list-group-item list-group-item-action" data-toggle="collapse" data-target="#addgraph"></a>').text(title)
+        const title = graphs[graph].graph_title || graph.split('/')[2]
+        const graphelement = $('<a href="#" class="list-group-item list-group-item-action" data-toggle="collapse" data-target="#addgraph"></a>').text(title)
         // add the host graph unless a host graph has been selected
         if (!hostFilter) {
           graphelement.prepend($('<small></small>').text(graph.split('/')[1] + ' / '))
@@ -632,6 +590,7 @@ $(document).ready(function () {
           })
         }
       }
+      return undefined
     }).toArray()
   }
 
@@ -656,7 +615,7 @@ $(document).ready(function () {
     clearGraphs()
     graphs.forEach(function (graph) {
       // lookup the graph by name
-      var plot = addGraph(document.graph_data[graph.name], graph.size || 'sm')
+      const plot = addGraph(document.graph_data[graph.name], graph.size || 'sm')
       // hide fields
       if (graph.hidden && graph.hidden.length) {
         graph.hidden.forEach(function (field) {
@@ -687,15 +646,21 @@ $(document).ready(function () {
       $('#dashboards').remove()
     } else {
       Object.keys(dashboards).sort().forEach(function (name) {
-        var option = $('<button class="dropdown-item" type="button">').text(name)
+        const option = $('<button class="dropdown-item" type="button">').text(name)
         $('#dashboards .dropdown-menu').append(option)
         option.click(function () {
-          var dashboard = dashboards[name]
+          const dashboard = dashboards[name]
           setGraphs(dashboard.graphs)
-          setDateRange(dashboard.dateRange.start, dashboard.dateRange.end)
+          if (dashboard.dateRange) {
+            setDateRange(dashboard.dateRange.start, dashboard.dateRange.end)
+          }
           $('#dashboards button.dropdown-toggle span').text(name)
         })
       })
+      // add import action
+      $('#dashboards .dropdown-menu').append('<div class="dropdown-divider"></div>')
+      $('#loadDashboardBtn').removeClass('btn btn-outline-secondary').addClass('dropdown-item')
+      $('#dashboards .dropdown-menu').append($('#loadDashboardBtn').append(' Load'))
     }
   })
 
@@ -712,4 +677,108 @@ $(document).ready(function () {
       setGraphs(JSON.parse(localStorage.getItem('shownGraphs')))
     } catch (error) {}
   })
+
+  // handle showing and hiding of the save dashboard dialog
+  $('#saveDashboard').on('shown.bs.modal', function () {
+    $('#saveDashboardName').trigger('focus')
+  })
+  $('#saveDashboard').on('hidden.bs.modal', function () {
+    $('#saveDashboardName').val('')
+    $('#saveDashboard .alert').remove()
+  })
+
+  // output readable but compact JSON
+  function getSaveDashboardData() {
+    let value = '{\n'
+    const name = $('#saveDashboardName').val()
+    if (name) {
+      value += '  "name": ' + JSON.stringify(name) + ',\n'
+    }
+    if ($('#saveDashboardDateRange:checked').length) {
+      const dateRange = JSON.parse(localStorage.getItem('dateRange'))
+      value += '  "dateRange": ' + JSON.stringify(dateRange) + ',\n'
+    }
+    const graphs = JSON.parse(localStorage.getItem('shownGraphs'))
+    value += '  "graphs": ' + JSON.stringify(graphs) + '\n}'
+    return value
+  }
+
+  // save dialog copy data to clipboard
+  $('#saveDashboardClipboard').on('click', function (event) {
+    // create a temporary text area
+    const textarea = document.createElement('textarea')
+    $(textarea).css({
+      position: 'absolute',
+      left: '-1000px',
+      top: '-1000px'
+    })
+    // fill text area with JSON
+    textarea.value = getSaveDashboardData()
+    // copy text area to clipboard
+    $('#saveDashboard form').append(textarea)
+    textarea.select()
+    textarea.setSelectionRange(0, 99999)
+    document.execCommand('copy')
+    $('#saveDashboard form textarea').remove()
+    // show notification
+    $('<div class="alert alert-success">Copied to clipboard</div>').hide().appendTo('#saveDashboard .modal-body').show(200, function () {
+      setTimeout(function () {
+        $('#saveDashboard').modal('hide')
+      }, 600)
+    })
+  })
+
+  // save dashboard definition to file
+  $('#saveDashboardFile').on('click', function (event) {
+    const element = document.createElement('a')
+    element.setAttribute('href', 'data:application/json;charset=utf-8,' + encodeURIComponent(getSaveDashboardData() + '\n'))
+    element.setAttribute('download', 'dashboard.json')
+    element.style.display = 'none'
+    document.body.appendChild(element)
+    element.click()
+    document.body.removeChild(element)
+    $('#saveDashboard').modal('hide')
+  })
+
+  // handle showing and hiding of the load dashboard dialog
+  $('#loadDashboard').on('shown.bs.modal', function () {
+    $('#loadDashboardData').trigger('focus')
+  })
+  $('#loadDashboard').on('hidden.bs.modal', function () {
+    $('#loadDashboardData').val('')
+    $('#loadDashboard .alert').remove()
+  })
+
+  // load dashboard definition from file into text area
+  $('#loadDashboardFileBtn').on('click', function (event) {
+    $('#loadDashboardFile').click()
+  })
+  $('#loadDashboardFile').on('change', function (event) {
+    const reader = new FileReader()
+    reader.onload = function (e) {
+      $('#loadDashboardData').val(e.target.result)
+    }
+    reader.readAsText(this.files[0], 'utf-8')
+    $(this).val('')
+  })
+
+  // load dashboard
+  $('#loadDashboardGo').on('click', function (event) {
+    try {
+      const dashboard = JSON.parse($('#loadDashboardData').val())
+      if (dashboard) {
+        setGraphs(dashboard.graphs)
+        if (dashboard.dateRange) {
+          setDateRange(dashboard.dateRange.start, dashboard.dateRange.end)
+        }
+        if (dashboard.name) {
+          $('#dashboards button.dropdown-toggle span').text(dashboard.name)
+        }
+        $('#loadDashboard').modal('hide')
+      }
+    } catch (e) {
+      $('#loadDashboard .alert').remove()
+      $('<div class="alert alert-danger"></div>').text('Error: ' + e.name + ': ' + e.message).hide().appendTo('#loadDashboard .modal-body').show(200)
+    }
+  })
 })
